#!/usr/bin/python3

#  This is free and unencumbered software released into the public domain.
#
#  Anyone is free to copy, modify, publish, use, compile, sell, or
#  distribute this software, either in source code form or as a compiled
#  binary, for any purpose, commercial or non-commercial, and by any
#  means.
#
#  In jurisdictions that recognize copyright laws, the author or authors
#  of this software dedicate any and all copyright interest in the
#  software to the public domain. We make this dedication for the benefit
#  of the public at large and to the detriment of our heirs and
#  successors. We intend this dedication to be an overt act of
#  relinquishment in perpetuity of all present and future rights to this
#  software under copyright law.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
#  OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.
#
#  For more information, please refer to <https://unlicense.org>

# NOTE
# 使用示例:
#   python3 main.py n8680fz
#
# 依赖 requests 库，请按照 https://requests.readthedocs.io/en/latest/user/install/#install 安装
# 在Linux使用时记得命令行配置 Proxy 环境变量

## 配置项

# 小说家站点URL，无尾随斜杠
c_base_url = "https://ncode.syosetu.com"
# 请求的头部
c_headers = {
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64; rv:102.0) Gecko/20100101 Firefox/102.0"
}
# 是否在文件名之前加入章节序号，关闭此选项可能导致重名问题
c_add_index = True
# 存储为文件时章节序号和标题之间的分隔符
c_separator = '-'
# 存储为文件时章节序号的前导零填充到的位数，设为0可以移除前导零
c_pad_head_count = 4
# 存储为文件时的后缀，可选:
#   ".html":    全页面抓取
#   ".md":      仅抓取正文(推荐方法)
#   ".txt":     抓取正文并剥离HTML标签(可能存在问题)
c_file_extension = ".md"

import sys
import requests
from html.parser import HTMLParser
from os.path import exists
from os import makedirs
import re

# .md 和 .txt 后缀模式时使用的匹配表达式，获取第一个捕获括号
c_regexp = re.compile("\<div id\=\"novel_honbun\".*?\>(.*?)\<\/div\>", re.DOTALL)

## 这Python的HTML解析器是真滴难用，什么科学的设计，我真的谢谢。。。

def attr(attrs, name):
    for (key, value) in attrs:
        if key == name:
            return value
    return ''
def klass(attrs, klass_name):
    return klass_name in attr(attrs, "class").split(' ')

class IndexParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.results = []
        self.matched_novel_sublist = False
        self.matched_a = False

    def handle_starttag(self, tag, attrs):
        if klass(attrs, "novel_sublist2"):
            self.results.append(['', ''])
            self.matched_novel_sublist = True
            self.matched_a = False
        elif self.matched_novel_sublist and not self.matched_a and tag == 'a':
            self.results[-1][0] = attr(attrs, "href")
            self.matched_a = True

    def handle_data(self, data):
        if self.matched_a:
            self.results[-1][1] += data

    def handle_endtag(self, tag):
        if self.matched_a:
            self.matched_a = False
            self.matched_novel_sublist = False

class ContentParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.results = []

    def handle_starttag(self, tag, attrs):
        if tag == "br":
            self.results.append('\n')

    def handle_data(self, data):
        self.results.append(data)

    def get_content(self, html):
        self.results = []
        self.reset()
        self.feed(html)
        self.close()
        return ''.join(self.results)

novel_id = ''
if len(sys.argv) > 1:
    novel_id = sys.argv[1]

if not novel_id:
    print("请输入小说ID，示例：")
    print("  $ python3 main.py n8680fz")
    sys.exit()

print("[info] fetching index...")

index_req = requests.get(c_base_url + '/' + novel_id, headers=c_headers)

if index_req.status_code == 200:
    parser = IndexParser()
    content_parser = ContentParser()
    parser.feed(index_req.text)
    index = parser.results

    print("[info] index fetched:", len(index), "sections")

    makedirs(novel_id, mode=0o777, exist_ok=True)
    for [i, [href, title]] in enumerate(index):
        nid = href.rsplit('/', 2)[1].zfill(c_pad_head_count)
        file_name = ((nid + c_separator) if c_add_index else '') + str.strip(title)
        file_path = novel_id + '/' + file_name + c_file_extension

        print("[info] prepare to fetch:", file_name)

        if exists(file_path):
            print("[info] already fetched, skipping")
        else:

            print("[info] fetching...")
            page_req = requests.get(c_base_url + href, headers=c_headers)

            if page_req.status_code == 200:
                content = ''

                if c_file_extension == ".html":
                    content = page_req.text
                elif c_file_extension == ".md":
                    content = str.strip(c_regexp.search(page_req.text).group(1))
                elif c_file_extension == ".txt":
                    content = content_parser.get_content(str.strip(c_regexp.search(page_req.text).group(1)))
                else:
                    print("[error] unknown file extension")
                    break

                file = open(file_path, 'w', encoding="utf-8")
                file.write(content)
                file.close()
                print("[info] fetch complete")

            else:
                print("[error] failed to fetch")
                break

    print("[info] all complete")

else:
    print("[error] failed to fetch index page")
